/*******************************************************************************
 * $Id: Aligner.h,v 1.23 2004/04/19 23:10:13 darling Exp $
 * This file is copyright 2002-2004 Aaron Darling.  All rights reserved.
 * Please see the file called COPYING for licensing, copying, and modification
 * rights.  Redistribution of this file, in whole or in part is prohibited
 * without express permission.
 ******************************************************************************/

#ifndef _Aligner_h_
#define _Aligner_h_

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "libGenome/gnSequence.h"
#include "libMems/DNAMemorySML.h"
//#include "DNAFileSML.h"
#include "libMems/MatchList.h"
#include "libMems/Interval.h"
#include "libMems/IntervalList.h"
#include "libMems/MemHash.h"
#include "libMems/MaskedMemHash.h"
#include <map>
#include "libMems/NumericMatrix.h"
#include <list>
#include "libMems/SlotAllocator.h"


/**
 * A mem labeled with a number.
 * Used by LCB construction algorithm 
 */
class LabeledMem{
public:
	Match* mem;
	uint label;
};

/**
 * Compares Matches labeled with a number.
 * Used by LCB construction algorithm 
 */
class LabeledMemComparator {
public:
	LabeledMemComparator( uint seq ){
		m_seq = seq;
	}
	LabeledMemComparator( LabeledMemComparator& lmc ){
		m_seq = lmc.m_seq;
	}
	boolean operator()(const LabeledMem& a, const LabeledMem& b) const{
		
		int64 a_start = a.mem->Start( m_seq ), b_start = b.mem->Start( m_seq );
		if( a_start == MEM_NO_MATCH || b_start == MEM_NO_MATCH ){
			if( b_start != MEM_NO_MATCH )
				return true;
			return false;
		}
		if(a_start < 0)
			a_start = -a_start;
//			a_start = -a_start + a.mem->Length();
		if(b_start < 0)
			b_start = -b_start;
//			b_start = -b_start + b.mem->Length();
		int64 diff = a_start - b_start;
		return diff < 0;
	}
	~LabeledMemComparator() {
	}
protected:
	uint m_seq;
private:
	LabeledMemComparator();
};

/**
 * A match with an associated list iterator.
 * Used by LCB construction algorithm 
 */
class PlacementMatch{
public:
	Match* mem;
	list< LabeledMem >::iterator iter;
};

/**
 * Compares Matches.
 * Used by LCB construction algorithm 
 */
class PlacementMatchComparator {
public:
	PlacementMatchComparator( uint seq ){
		m_seq = seq;
	}
	PlacementMatchComparator( PlacementMatchComparator& lmc ){
		m_seq = lmc.m_seq;
	}
	boolean operator()(const PlacementMatch& a, const PlacementMatch& b) const{
		
		int64 a_start = a.mem->Start( m_seq ), b_start = b.mem->Start( m_seq );
		if( a_start == MEM_NO_MATCH || b_start == MEM_NO_MATCH ){
			if( b_start != MEM_NO_MATCH )
				return true;
			return false;
		}
		if(a_start < 0)
			a_start = -a_start;
//			a_start = -a_start + a.mem->Length();
		if(b_start < 0)
			b_start = -b_start;
//			b_start = -b_start + b.mem->Length();

		int64 diff = a_start - b_start;
		return diff < 0;
	}
	~PlacementMatchComparator() {
	}
protected:
	uint m_seq;
private:
	PlacementMatchComparator();
};



/** 
 * This class is used to track relationships between LCBs during the LCB determination process.
 */
class LCB{
public:
//	vector< int64 > start;		/**< The start position of the LCB in each sequence */
//	vector< int64 > length;		/**< The length of the LCB in each sequence */
	vector< int64 > left_end;	/**< The left end position of the LCB in each sequence */
	vector< int64 > right_end;  /**< The right end position of the LCB in each sequence */
	vector< uint > left_adjacency;	/**< 'Pointers' (actually IDs) to the LCBs on the left in each sequence */
	vector< uint > right_adjacency;	/**< 'Pointers' (actually IDs) to the LCBs on the right in each sequence */
	int lcb_id;			/**< A numerical ID that can be assigned to this LCB */
	int64 weight;		/**< The weight (or coverage) of this LCB */
	boolean recursed_left;	/**< Flag for wether the left side of this LCB has been searched recursively */
	boolean recursed_right;	/**< Flag for wether the right side of this LCB has been searched recursively */
};

/**
 * Used to find locally colinear blocks (LCBs) and do recursive
 * alignments on the blocks
 * To create an alignment one need only use the align method.
 * LCB lists are typically stored using the IntervalList class.  They can be
 * read and written in interval format using that class.  For input and output
 * of gapped alignments in other formats, see the gnAlignedSequences class.
 * Other methods in this class are available for experimentation.
 */
class Aligner {
public:
	/** 
	 * Constructs an aligner for the specified number of sequences.
	 * @param seq_count 	The number of sequences that will be aligned with this Aligner
	 */
	Aligner( uint seq_count );
	Aligner( const Aligner& al );
	Aligner& operator=( const Aligner& al );

	/**
	 * Performs an alignment.  Takes a MatchList as input and outputs a list of LCBs as an IntervalList.
	 * Several of the options can be used to filter out unlikely LCBs.  If the recursive option is
	 * specified, the regions between matches in each LCB are searched for further homology and a full
	 * gapped alignment is produced.
	 * @param mlist					The MatchList to use as input for the alignment process
	 * @param interval_list			The IntervalList that is created by the alignment process
	 * @param LCB_minimum_density	The minimum density that an LCB may have to be considered a valid block
	 *   						 	This should be a number between 0 and 1.
	 * @param LCB_minimum_range		A misnomer: really it's the minimum number of matching base pairs an LCB 
	 *								must contain to be considered an LCB. Coverage is defined as 
	 *								(length of match) * (# of matching sequences)
	 * @param recursive 			Option for performing a recursive alignment.  If this is set to
	 * 								true, all regions which have gaps will be searched for exact matches.
	 * @param extend_lcbs		If true, attempt to extend the boundaries of LCBs by searching for 
	 *                          additional matches between LCBs
	 * @param tree_filename		The name of the output file to write the phylogenetic guide tree into.  If
	 *                          an empty string is specified then a temporary file is created.  
	 * @throws AlignerError 	may be thrown if an error occurs
	 */
	void align( MatchList& mlist, IntervalList& interval_list, double LCB_minimum_density, double LCB_minimum_range, boolean recursive, boolean extend_lcbs, boolean gapped_alignment, string tree_filename = "" );
	
	/**
	 * Function to determine the breakpoints in a set of matches. 
	 * Sorts the matches in mlist and returns the indices of breakpoints.
	 * This function attempts (sometimes unsuccessfully) to determine subset LCBs.  If a set of
	 * matches containing subset LCBs has been passed to it, the resulting breakpoint set may
	 * be incorrect.  You have been warned.
	 * @param mlist 		A list of matches to search for LCBs.
	 * @param breakpoints 	The indices of matches in the sorted match list that are at LCB boundaries
	 */
	void AaronsLCB( MatchList& mlist, set<uint>& breakpoints );

	/**
	 * Scans a set of numbers for inclusion of each successive number starting from a particular value
	 * and either increasing or decreasing from that value.  Used by AaronsLCB when scanning past subset
	 * matches.
	 */
	void scanLabels( set< uint >& no_match_labels, uint& start_label, boolean forward );
	/**
	 * Compares to (subset) matches to determine if they are collinear in every sequence they are
	 * both defined in.
	 * @param m1 	A Match
	 * @param m2 	A Match
	 * @return 		true if m1 and m2 are collinear, false otherwise.
	 */
	boolean checkCollinearity( Match* m1, Match* m2 );
	void scanFit( list< LabeledMem >& pair_list, list< LabeledMem >::iterator& list_iter, Match* new_match, uint sort_seq );

	void Recursion( MatchList& r_list, Match* r_begin, Match* r_end, boolean nway_only = false );
	void GetBestLCB( MatchList& r_list, MatchList& best_lcb );
	void DoSomethingCool( MatchList& mlist, Interval& iv );
	
	/**
	 * Deletes overlapping regions in a set of matches.  Always removes matching base pairs from the
	 * match covering fewer bases.  Coverage is defined as (length of match) * (# of matching sequences)
	 */
	static void EliminateOverlaps( MatchList& ml );
		
	
	void GetLCBCoverage( MatchList& lcb, uint64& coverage );
	void ComputeLCBs( MatchList& meml, set<uint>& breakpoints, vector<MatchList>& lcb_list, vector<int64>& weights );
	void computeLCBAdjacencies_v2( vector<MatchList>& lcb_list, vector< int64 >& weights, vector< LCB >& adjacencies );
	void computeLCBAdjacencies_v2( IntervalList& iv_list, vector< int64 >& weights, vector< LCB >& adjacencies );
	void scanLeft( int& left_recurseI, vector< LCB >& adjacencies, int min_weight, int seqI );
	void scanRight( int& right_recurseI, vector< LCB >& adjacencies, int min_weight, int seqI );
	
	void SearchLCBGaps( vector< LCB >& adjacencies, MatchList& new_matches, boolean entire_genome );
	/**
	 * Set the minimum size of intervening region between two anchor matches that will
	 * be considered for recursive anchor determination.  When the gaps between two anchors
	 * are less than this cutoff value the region is handed off to the dynamic programming aligner
	 * e.g. ClustalW 
	 */
	void SetMinRecursionGapLength( gnSeqI min_r_gap );

	void greedyBreakpointElimination( gnSeqI minimum_weight, vector< LCB >& adjacencies, vector< MatchList >& lcb_list, vector< int64 >& weights, ostream* status_out = NULL );
	void filterMatches( vector< LCB >& adjacencies, vector< MatchList >& lcb_list, vector< int64 >& weights );


	void SearchWithinLCB( MatchList& mlist );
	void RecursiveAnchorSearch( MatchList& mlist, gnSeqI minimum_weight, vector< MatchList >& LCB_list, boolean entire_genome, ostream* status_out = NULL );

	void AlignLCB( MatchList& mlist, Interval& iv );

protected:
	MatchList ml;
	MemHash gap_mh;			/**< Used during recursive alignment */
	MaskedMemHash nway_mh;	/**< Used during recursive alignment to find nway matches only */
	uint32 seq_count;		/**< The number of sequences this aligner is working with */
	boolean debug;			/**< Flag for debugging output */
	
	double LCB_minimum_density;
	double LCB_minimum_range;
	
	int64 cur_min_coverage;	/**< Tracks the minimum weight of the least weight LCB */
	
	gnSeqI min_recursive_gap_length;	/**< Minimum size of gap regions that will be recursed on */
	SlotAllocator< Match >& match_allocator;

	void consistencyCheck( uint lcb_count, vector< LCB >& adjacencies, vector< MatchList >& lcb_list, vector< int64 >& weights );
	
	boolean recursive;		/**< Set to true if a recursive anchor search/gapped alignment should be performed */
	boolean extend_lcbs;	/**< Set to true if LCB extension should be attempted */
	boolean gapped_alignment;	/**< Set to true to complete a gapped alignment */
	boolean currently_recursing;	/**< True when the recursive search has begun */
	boolean collinear_genomes;	/**< Set to true if all genomes are assumed to be collinear */

};

/**
 * Thrown if some error occurs during alignment
 */
CREATE_EXCEPTION( AlignerError );

void transposeMatches( MatchList& mlist, uint seqI, const vector< int64 >& seq_regions );

#endif // _Aligner_h_
